@@ -2,18 +2,18 @@ import json
 import os
 import sys
 import urllib.parse
-from datetime import datetime
+from datetime import date, datetime, timedelta
 from re import match, search
 from typing import Any, Dict, Optional, Set, Tuple
 
 from locust.clients import HttpSession, ResponseContextManager
 from requests import Response
 
-from ._save_request_builder import save_builder
-from .helper import (find_component_by_attribute_in_dict, list_filter,
-                     log_locust_error, test_response_for_error, get_username)
 from . import logger
+from ._save_request_builder import save_builder
 from .exceptions import BadCredentialsException, MissingCsrfTokenException
+from .helper import (find_component_by_attribute_in_dict, get_username,
+                     log_locust_error, test_response_for_error)
 
 log = logger.getLogger(__name__)
 
@@ -169,6 +169,7 @@ class _Interactor:
         # load initial page to get tokens/cookies
         token_uri = uri + '?signin=native'
         resp = self.get_page(token_uri, label="Login.LoadUi", check_login=False)
+        log.info(f"Attempting to load page {self.replace_base_path_if_appropriate(token_uri)}")
         payload = {
             "un": self.auth[0],
             "pw": self.auth[1],
@@ -184,10 +185,11 @@ class _Interactor:
             "Referer": self.host,
             "User-Agent": "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_12_6) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/61.0.3163.79 Safari/537.36",
         }
-
+        login_url = uri + "auth?appian_environment=tempo"
+        log.info(f"Logging in at {self.replace_base_path_if_appropriate(login_url)}")
         # Post Auth
         resp = self.post_page(
-            uri + "auth?appian_environment=tempo",
+            login_url,
             payload=urllib.parse.urlencode(payload),
             headers=headers,
             label="Login.SubmitAuth",
@@ -318,6 +320,8 @@ class _Interactor:
             proposed_file_name = proposed_file_name[:-length_of_file_type] + " (1)" + file_ending
         with open(proposed_file_name, 'w') as f:
             f.write(response.text)
+        if 'X-Trace-Id' in response.headers:
+            log.info(cleaned_label + ' | X-Trace-Id: ' + response.headers['X-Trace-Id'])
 
     def click_record_link(self, get_url: str, component: Dict[str, Any], context: Dict[str, Any],
                           label: str = None, headers: Dict[str, Any] = None, locust_label: str = "") -> Dict[str, Any]:
@@ -408,7 +412,7 @@ class _Interactor:
         )
         return resp.json()
 
-    def click_start_process_link(self, label: str, component: Dict[str, Any], process_model_opaque_id: str,
+    def click_start_process_link(self, component: Dict[str, Any], process_model_opaque_id: str,
                                  cache_key: str, site_name: str, page_name: str, is_mobile: bool = False,
                                  locust_request_label: str = None) -> Dict[str, Any]:
         '''
@@ -430,10 +434,7 @@ class _Interactor:
         else:
             spl_link_url = f"/suite/rest/a/sites/latest/{site_name}/page/{page_name}/startProcess/{process_model_opaque_id}?cacheKey={cache_key}"
 
-        headers = self.setup_request_headers()
-        headers["Accept"] = "application/vnd.appian.tv.ui+json"
-        headers["Content-Type"] = "application/vnd.appian.tv+json"
-
+        headers = self.setup_sail_headers()
         locust_label = locust_request_label or "Clicking StartProcessLink: " + component["label"]
         resp = self.post_page(
             self.host + spl_link_url, payload={}, headers=headers, label=locust_label
@@ -458,10 +459,7 @@ class _Interactor:
         # Web url:
         related_action_link_url = f"/suite/rest/a/record/latest/{record_type_stub}/{opaque_record_id}/actions/{opaque_related_action_id}"
 
-        headers = self.setup_request_headers()
-        headers["Accept"] = "application/vnd.appian.tv.ui+json"
-        headers["Content-Type"] = "application/vnd.appian.tv+json"
-
+        headers = self.setup_sail_headers()
         locust_label = locust_request_label or "Clicking RelatedActionLink: " + component["label"]
         resp = self.post_page(
             self.host + related_action_link_url, payload={}, headers=headers, label=locust_label
@@ -784,10 +782,10 @@ class _Interactor:
 
             Args:
                 post_url: the url (not including the host and domain) to post to
-                checkbox: the JSON representing the desired checkbox
+                upload_field: the JSON representing the desired checkbox
                 context: the Sail context parsed from the json response
                 uuid: the uuid parsed from the json response
-                indices: indices of the checkbox
+                doc_id: document id for the upload
                 context_label: the label to be displayed by locust for this action
                 client_mode: where this is being uploaded to, defaults to DESIGN
 
@@ -814,6 +812,81 @@ class _Interactor:
         )
         return resp.json()
 
+    def update_date_field(self, post_url: str, date_field_component: Dict[str, Any],
+                          date_input: date, context: Dict[str, Any], uuid: str,
+                          locust_label: str = None) -> Dict[str, Any]:
+        '''
+            Calls the post operation to update a date field
+
+            Args:
+                post_url: the url (not including the host and domain) to post to
+                date_field_component: the JSON representing the date field component
+                date_input: date field to convert to proper text format
+                context: the Sail context parsed from the json response
+                uuid: the uuid parsed from the json response
+
+            Returns: the response of post operation as json
+        '''
+        new_value = {
+            "#t": "date",
+            "#v": f"{date_input.isoformat()}Z" if date_input else None
+        }
+
+        payload = save_builder()\
+            .component(date_field_component)\
+            .context(context)\
+            .uuid(uuid)\
+            .value(new_value)\
+            .build()
+
+        locust_label = locust_label or "Filling Date Field for " + \
+            date_field_component.get("label", date_field_component.get("testLabel", "DateField"))
+
+        headers = self.setup_sail_headers()
+
+        resp = self.post_page(
+            self.host + post_url, payload=payload, headers=headers, label=locust_label
+        )
+        return resp.json()
+
+    def update_datetime_field(self, post_url: str, datetime_field: Dict[str, Any],
+                              datetime_input: datetime, context: Dict[str, Any], uuid: str,
+                              locust_label: str = None) -> Dict[str, Any]:
+        '''
+            Calls the post operation to update a date field
+
+            Args:
+                post_url: the url (not including the host and domain) to post to
+                datetime_field: the JSON representing the datetime field to edit
+                datetime_input: datetime field to convert to the proper text format
+                context: the Sail context parsed from the json response
+                uuid: the uuid parsed from the json response
+
+            Returns: the response of post operation as json
+        '''
+
+        new_value = {
+            "#t": "dateTime",
+            "#v": f"{datetime_input.replace(second=0, microsecond=0).isoformat()}Z" if datetime_input else None
+        }
+
+        payload = save_builder()\
+            .component(datetime_field)\
+            .context(context)\
+            .uuid(uuid)\
+            .value(new_value)\
+            .build()
+
+        locust_label = locust_label or "Filling Date Time Field for " + \
+            datetime_field.get("label", datetime_field.get("testLabel", "DateField"))
+
+        headers = self.setup_sail_headers()
+
+        resp = self.post_page(
+            self.host + post_url, payload=payload, headers=headers, label=locust_label
+        )
+        return resp.json()
+
     def update_grid_from_sail_form(self, post_url: str,
                                    grid_component: Dict[str, Any], new_grid_save_value: Dict[str, Any],
                                    context: Dict[str, Any], uuid: str,
