@@ -1,3 +1,4 @@
+import datetime
 import enum
 import errno
 import json
@@ -6,7 +7,7 @@ import random
 import warnings
 from functools import wraps
 from typing import Any, Callable, Dict, List, Optional
-from urllib.parse import urlparse
+from urllib.parse import quote, urlparse
 
 from appian_locust.records_helper import _is_grid
 
@@ -14,8 +15,7 @@ from . import logger
 from ._grid_interactor import GridInteractor
 from ._interactor import _Interactor
 from ._ui_reconciler import UiReconciler
-from .helper import (extract_all_by_label, extract_item_by_label,
-                     find_component_by_attribute_in_dict,
+from .helper import (extract_all_by_label, find_component_by_attribute_in_dict,
                      find_component_by_index_in_dict,
                      find_component_by_label_and_type_dict, log_locust_error)
 from .records_helper import (get_record_header_response,
@@ -141,7 +141,7 @@ class SailUiForm:
         if not new_state:
             raise Exception(f"No response returned when trying to update field with label '{label}'")
 
-        return self._reconcile_and_produce_new_form(new_state, form_url=reeval_url)
+        return self._reconcile_state(new_state, form_url=reeval_url)
 
     @raises_locust_error("uiform.py/fill_field_by_index()")
     def fill_field_by_index(self, type_of_component: str, index: int, text_to_fill: str, locust_request_label: str = "") -> 'SailUiForm':
@@ -176,7 +176,7 @@ class SailUiForm:
                     with index: '{index}' on the current page")
                 ''')
 
-        return self._reconcile_and_produce_new_form(new_state, form_url=reeval_url)
+        return self._reconcile_state(new_state, form_url=reeval_url)
 
     @raises_locust_error("uiform.py/fill_field_by_any_attribute()")
     def fill_field_by_any_attribute(self, attribute: str, value_for_attribute: str, text_to_fill: str, locust_request_label: str = "") -> 'SailUiForm':
@@ -215,7 +215,7 @@ class SailUiForm:
         reeval_url = self._get_update_url_for_reeval(self.state)
         locust_label = locust_request_label or f"{self.breadcrumb}.FillTextFieldByAttribute.{attribute}"
         new_state = self.interactor.fill_textfield(
-            reeval_url, component, text_to_fill, self.context, self.uuid, label=locust_request_label)
+            reeval_url, component, text_to_fill, self.context, self.uuid, label=locust_label)
         if not new_state:
             raise Exception(
                 f'''
@@ -223,7 +223,7 @@ class SailUiForm:
                     and its value: '{value_for_attribute}' on the current page")
                 ''')
 
-        return self._reconcile_and_produce_new_form(new_state, form_url=reeval_url)
+        return self._reconcile_state(new_state, form_url=reeval_url)
 
     # Aliases for fill_text_field() function
     fill_paragraph_field = fill_text_field
@@ -296,12 +296,12 @@ class SailUiForm:
         if not newer_state:
             raise Exception(f"No response returned when trying to update field with label '{label}'")
 
-        return self._reconcile_and_produce_new_form(new_state)
+        return self._reconcile_state(new_state)
 
     @raises_locust_error("uiform.py/click()")
     def click(self, label: str, is_test_label: bool = False, locust_request_label: str = "") -> 'SailUiForm':
         """
-        Fills a field on the form, if there is one present with the following label (case sensitive)
+        Clicks on a component on the form, if there is one present with the following label (case sensitive)
         Otherwise throws a NotFoundException
 
         Args:
@@ -331,7 +331,7 @@ class SailUiForm:
 
         if not new_state:
             raise Exception(f"No response returned when trying to click button with label '{label}'")
-        return self._reconcile_and_produce_new_form(new_state, form_url=reeval_url)
+        return self._reconcile_state(new_state, form_url=reeval_url)
 
     # Aliases for click() function
     click_button = click
@@ -340,10 +340,13 @@ class SailUiForm:
     @raises_locust_error("uiform.py/click_card_layout_by_index()")
     def click_card_layout_by_index(self, index: int, locust_request_label: str = "") -> 'SailUiForm':
         """
-        Clicks a card layout link by index
+        Clicks a card layout link by index.
+        This method will find the CardLayout component on the UI by index and then perform
+        the click behavior on its Link component.
 
         Args:
-            index(int): Index of the card layout on which to click.
+            index(int): Index of the card layout on which to click. (first found card layout , or second etc.)
+                        (Indices start from 1)
 
         Keyword Args:
             locust_request_label(str): Label used to identify the request for locust statistics
@@ -351,21 +354,33 @@ class SailUiForm:
         Returns (SailUiForm): The latest state of the UiForm
 
         Examples:
+            This finds link field on the 2nd card layout on the page and clicks it. It handles StartProcessLink as well, so
+            no need to call click_start_process_link() when it is on a CardLayout.
 
             >>> form.click_card_layout_by_index(2)
 
         """
         component = find_component_by_index_in_dict("CardLayout", index, self.state)
-        # Add an arbitrary label to the cardlayout component to help with link clicking
-        component["label"] = "CardLayoutLabel"
+
+        if "link" in component and component["link"]:
+            link_component = component["link"]
+        else:
+            raise Exception(f"CardLayout found at index: {index} does not have a link on it")
+
         locust_label = locust_request_label or f"{self.breadcrumb}.ClickCardLayout.Index.{index}"
-        reeval_url = self._get_update_url_for_reeval(self.state)
-        new_state = self.interactor.click_component(
-            reeval_url, component, self.context, self.uuid, label=locust_label)
+
+        if link_component["#t"] == "StartProcessLink":
+            site_name = link_component["siteUrlStub"] or "D6JMim"
+            page_name = link_component["sitePageUrlStub"]
+            new_state = self._click_start_process_link(site_name, page_name, False, link_component, locust_request_label=locust_label)
+        else:
+            new_state = self.interactor.click_component(self.form_url, link_component, self.context, self.uuid, label=locust_label)
 
         if not new_state:
             raise Exception(f"No response returned when trying to click card layout at index '{index}'")
-        return self._reconcile_and_produce_new_form(new_state, form_url=reeval_url)
+
+        reeval_url = self._get_update_url_for_reeval(new_state)
+        return self._reconcile_state(new_state, form_url=reeval_url)
 
     @raises_locust_error("uiform.py/click_record_link()")
     def click_record_link(self, label: str, locust_request_label: str = "") -> 'SailUiForm':
@@ -389,7 +404,7 @@ class SailUiForm:
         locust_label = locust_request_label or f"{self.breadcrumb}.ClickRecordLink.{label}"
         new_state = self.interactor.click_record_link(reeval_url, component, self.context, self.uuid,
                                                       locust_label=locust_label)
-        return self._reconcile_and_produce_new_form(new_state, form_url=reeval_url)
+        return self._reconcile_state(new_state, form_url=reeval_url)
 
     @raises_locust_error("uiform.py/click_start_process_link()")
     def click_start_process_link(self, label: str, site_name: str, page_name: str, is_mobile: bool = False, locust_request_label: str = "") -> 'SailUiForm':
@@ -416,20 +431,12 @@ class SailUiForm:
         component = find_component_by_label_and_type_dict('label', label, 'StartProcessLink', self.state)
         self._validate_component_found(component, label)
 
-        process_model_opaque_id = component.get("processModelOpaqueId", "")
-        cache_key = component.get("cacheKey", "")
-
-        if not process_model_opaque_id or not cache_key:
-            raise Exception(f'''
-                            StartProcessLink component does not have process model opaque id or cache key set.
-                            ''')
         locust_label = locust_request_label or f"{self.breadcrumb}.ClickStartProcessLink.{label}"
-        new_state = self.interactor.click_start_process_link(label, component, process_model_opaque_id,
-                                                             cache_key, site_name, page_name, is_mobile,
-                                                             locust_request_label=locust_label)
+        new_state = self._click_start_process_link(site_name, page_name, is_mobile, component, locust_request_label=locust_label)
+
         # get the re-eval URI from links object of the response (new_state)
         reeval_url = self._get_update_url_for_reeval(new_state)
-        return self._reconcile_and_produce_new_form(new_state, form_url=reeval_url)
+        return self._reconcile_state(new_state, form_url=reeval_url)
 
     @raises_locust_error("uiform.py/click_start_process_link_on_mobile()")
     def click_start_process_link_on_mobile(self, label: str, site_name: str, page_name: str, locust_request_label: str = "") -> 'SailUiForm':
@@ -455,6 +462,23 @@ class SailUiForm:
         locust_label = locust_request_label or f"{self.breadcrumb}.ClickStartProcessLink.Mobile.{label}"
         return self.click_start_process_link(label, site_name, page_name, is_mobile=True, locust_request_label=locust_label)
 
+    def _click_start_process_link(self, site_name: str, page_name: str, is_mobile: bool,
+                                  component: Dict[str, Any], locust_request_label: str) -> Dict[str, Any]:
+        """
+        This internal function is called by both click_start_process_link() and click_card_layout_by_index().
+        It takes parameters needed for making a start process lnik call to the server and calls the interactor to do that.
+        """
+        process_model_opaque_id = component.get("processModelOpaqueId", "")
+        cache_key = component.get("cacheKey", "")
+        if not process_model_opaque_id:
+            raise Exception(f"StartProcessLink component does not have process model opaque id set.")
+        elif not cache_key:
+            raise Exception(f"StartProcessLink component does not have cache key set.")
+
+        return self.interactor.click_start_process_link(component=component, process_model_opaque_id=process_model_opaque_id,
+                                                        cache_key=cache_key, site_name=site_name, page_name=page_name, is_mobile=is_mobile,
+                                                        locust_request_label=locust_request_label)
+
     @raises_locust_error("uiform.py/click_related_action_link()")
     def click_related_action(self, label: str, locust_request_label: str = "") -> 'SailUiForm':
         """
@@ -498,10 +522,10 @@ class SailUiForm:
         locust_label = locust_request_label or f"{self.breadcrumb}.ClickRelatedActionLink.{label}"
 
         new_state = self.interactor.click_related_action(component, record_type_stub, opaque_record_id,
-                                                         opaque_related_action_id, locust_request_label)
+                                                         opaque_related_action_id, locust_label)
         # get the re-eval URI from links object of the response (new_state)
         reeval_url = self._get_update_url_for_reeval(new_state)
-        return self._reconcile_and_produce_new_form(new_state, form_url=reeval_url)
+        return self._reconcile_state(new_state, form_url=reeval_url)
 
     @raises_locust_error("uiform.py/select_dropdown_item()")
     def select_dropdown_item(self, label: str, choice_label: str, locust_request_label: str = "", is_test_label: bool = False) -> 'SailUiForm':
@@ -548,7 +572,7 @@ class SailUiForm:
             raise Exception(
                 f"No response returned when trying to click button with label '{label}'")
 
-        return self._reconcile_and_produce_new_form(new_state, form_url=reeval_url)
+        return self._reconcile_state(new_state, form_url=reeval_url)
 
     def _check_checkbox_by_attribute(self, attribute: str, value_for_attribute: str, indices: List[int], locust_request_label: str = "") -> 'SailUiForm':
         """
@@ -578,7 +602,7 @@ class SailUiForm:
                             and its value: '{value_for_attribute}' on the current page
                             ''')
 
-        return self._reconcile_and_produce_new_form(new_state, form_url=reeval_url)
+        return self._reconcile_state(new_state, form_url=reeval_url)
 
     @raises_locust_error("uiform.py/check_checkbox_by_test_label()")
     def check_checkbox_by_test_label(self, test_label: str, indices: List[int], locust_request_label: str = "") -> 'SailUiForm':
@@ -605,7 +629,7 @@ class SailUiForm:
             raise Exception(f"No testLabel provided to select a checkbox")
 
         locust_label = locust_request_label or f'{self.breadcrumb}.CheckCheckboxByTestLabel.{test_label}'
-        return self._check_checkbox_by_attribute('testLabel', test_label, indices)
+        return self._check_checkbox_by_attribute('testLabel', test_label, indices, locust_request_label=locust_label)
 
     @raises_locust_error("uiform.py/check_checkbox_by_label()")
     def check_checkbox_by_label(self, label: str, indices: List[int], locust_request_label: str = "") -> 'SailUiForm':
@@ -664,7 +688,7 @@ class SailUiForm:
                 inside the TabButtonGroup component with testLabel: '{tab_group_test_label}'''
             )
 
-        return self._reconcile_and_produce_new_form(new_state, form_url=reeval_url)
+        return self._reconcile_state(new_state, form_url=reeval_url)
 
     @raises_locust_error("uiform.py/upload_document_to_upload_field()")
     def upload_document_to_upload_field(self, label: str, file_path: str, locust_request_label: str = "") -> 'SailUiForm':
@@ -712,7 +736,87 @@ class SailUiForm:
         if not new_state:
             raise Exception(
                 f"No response returned when trying to upload file to field '{label}'")
-        return self._reconcile_and_produce_new_form(new_state)
+        return self._reconcile_state(new_state)
+
+    @raises_locust_error("uiform.py/fill_date_field()")
+    def fill_date_field(self, label: str, date_input: datetime.date, locust_request_label: str = "") -> 'SailUiForm':
+        """
+        Fills a date field with the specified date
+
+        Args:
+            label(str): Label of the date field
+            date_input(date): Date used to fill the field
+
+        Keyword Args:
+            locust_request_label(str): Label used to identify the request for locust statistics
+
+        Returns (SailUiForm): The latest state of the UiForm
+
+        Examples:
+
+            >>> form.fill_date_field('Today', datetime.date.today())
+            >>> form.fill_date_field('Date of Birth', datetime.date(1992, 12, 30))
+
+        """
+        if not isinstance(date_input, datetime.date):
+            raise Exception("Input must be of type datetime.date")
+        field_type = 'DatePickerField'
+        date_field = find_component_by_label_and_type_dict('label', label, field_type, self.state)
+        self._validate_component_found(date_field, label, type=field_type)
+
+        locust_label = locust_request_label or f'{self.breadcrumb}.FillDateField'
+        reeval_url = self._get_update_url_for_reeval(self.state)
+
+        new_state = self.interactor.update_date_field(
+            reeval_url, date_field, date_input, self.context, self.uuid, locust_label=locust_label)
+        if not new_state:
+            raise Exception(f'''No response returned when trying to update: '{label}'
+                            with its value: '{date_input}' on the current page
+                            ''')
+
+        return self._reconcile_state(new_state, form_url=reeval_url)
+
+    @raises_locust_error("uiform.py/fill_datetime_field()")
+    def fill_datetime_field(self, label: str, datetime_input: datetime.datetime, locust_request_label: str = "") -> 'SailUiForm':
+        """
+        Fills a datetime field with the specified datetime
+
+        NOTE: this does one api call for both the date and time, whereas filling the elements on screen requires
+        two separate evaluations, one to fill the date field and one to fill the time field. This is the
+        way the request would look if one of the fields were already filled.
+
+        Args:
+            label(str): Label of the datetime field
+            datetime_input(date): Date time used to fill the field
+
+        Keyword Args:
+            locust_request_label(str): Label used to identify the request for locust statistics
+
+        Returns (SailUiForm): The latest state of the UiForm
+
+        Examples:
+
+            >>> form.fill_datetime_field('Now', datetime.datetime.now())
+            >>> form.fill_datetime_field('Date and Time of Birth', datetime.datetime(1992, 12, 30, 12, 30, 5))
+
+        """
+        if not isinstance(datetime_input, datetime.datetime):
+            raise Exception("Input must be of type datetime.datetime")
+        field_type = 'DateTimePickerField'
+        datetime_field = find_component_by_label_and_type_dict('label', label, field_type, self.state)
+        self._validate_component_found(datetime_field, label, type=field_type)
+
+        locust_label = locust_request_label or f'{self.breadcrumb}.FillDateTimeField'
+        reeval_url = self._get_update_url_for_reeval(self.state)
+
+        new_state = self.interactor.update_datetime_field(
+            reeval_url, datetime_field, datetime_input, self.context, self.uuid, locust_label=locust_label)
+        if not new_state:
+            raise Exception(f'''No response returned when trying to update: '{label}'
+                            with its value: '{datetime_input}' on the current page
+                            ''')
+
+        return self._reconcile_state(new_state, form_url=reeval_url)
 
     @raises_locust_error("uiform.py/move_to_end_of_paging_grid()")
     def move_to_end_of_paging_grid(self, label: str = None, index: int = None, locust_request_label: str = "") -> 'SailUiForm':
@@ -742,7 +846,7 @@ class SailUiForm:
         reeval_url = self._get_update_url_for_reeval(self.state)
         new_state = self.interactor.update_grid_from_sail_form(reeval_url, grid, new_grid_save,
                                                                self.context, self.uuid, context_label=context_label)
-        return self._reconcile_and_produce_new_form(new_state, form_url=reeval_url)
+        return self._reconcile_state(new_state, form_url=reeval_url)
 
     @raises_locust_error("uiform.py/move_to_beginning_of_paging_grid()")
     def move_to_beginning_of_paging_grid(self, label: str = None, index: int = None, locust_request_label: str = "") -> 'SailUiForm':
@@ -769,7 +873,7 @@ class SailUiForm:
         reeval_url = self._get_update_url_for_reeval(self.state)
         new_state = self.interactor.update_grid_from_sail_form(reeval_url, grid, new_grid_save,
                                                                self.context, self.uuid, context_label=context_label)
-        return self._reconcile_and_produce_new_form(new_state, form_url=reeval_url)
+        return self._reconcile_state(new_state, form_url=reeval_url)
 
     @raises_locust_error("uiform.py/move_to_left_in_paging_grid()")
     def move_to_left_in_paging_grid(self, label: str = None, index: int = None, locust_request_label: str = "") -> 'SailUiForm':
@@ -800,7 +904,7 @@ class SailUiForm:
         reeval_url = self._get_update_url_for_reeval(self.state)
         new_state = self.interactor.update_grid_from_sail_form(reeval_url, grid, new_grid_save,
                                                                self.context, self.uuid, context_label=context_label)
-        return self._reconcile_and_produce_new_form(new_state, form_url=reeval_url)
+        return self._reconcile_state(new_state, form_url=reeval_url)
 
     @raises_locust_error("uiform.py/move_to_right_in_paging_grid()")
     def move_to_right_in_paging_grid(self, label: str = None, index: int = None, locust_request_label: str = "") -> 'SailUiForm':
@@ -831,7 +935,7 @@ class SailUiForm:
         reeval_url = self._get_update_url_for_reeval(self.state)
         new_state = self.interactor.update_grid_from_sail_form(reeval_url, grid, new_grid_save,
                                                                self.context, self.uuid, context_label=context_label)
-        return self._reconcile_and_produce_new_form(new_state, form_url=reeval_url)
+        return self._reconcile_state(new_state, form_url=reeval_url)
 
     @raises_locust_error("uiform.py/sort_paging_grid()")
     def sort_paging_grid(self, label: str = None, index: int = None, field_name: str = "", ascending: bool = False, locust_request_label: str = "") -> 'SailUiForm':
@@ -869,7 +973,7 @@ class SailUiForm:
         reeval_url = self._get_update_url_for_reeval(self.state)
         new_state = self.interactor.update_grid_from_sail_form(reeval_url, grid, new_grid_save,
                                                                self.context, self.uuid, context_label=context_label)
-        return self._reconcile_and_produce_new_form(new_state, form_url=reeval_url)
+        return self._reconcile_state(new_state, form_url=reeval_url)
 
     @raises_locust_error("uiform.py/select_radio_button_by_test_label()")
     def select_radio_button_by_test_label(self, test_label: str, index: int, locust_request_label: str = "") -> 'SailUiForm':
@@ -900,7 +1004,7 @@ class SailUiForm:
         if not new_state:
             raise Exception(
                 f"No response returned when trying to select radio button with testLabel '{test_label}'")
-        return self._reconcile_and_produce_new_form(new_state, form_url=reeval_url)
+        return self._reconcile_state(new_state, form_url=reeval_url)
 
     @raises_locust_error("uiform.py/select_radio_button_by_label()")
     def select_radio_button_by_label(self, label: str, index: int, locust_request_label: str = "") -> 'SailUiForm':
@@ -934,7 +1038,7 @@ class SailUiForm:
         if not new_state:
             raise Exception(
                 f"No response returned when trying to select radio button with label '{label}'")
-        return self._reconcile_and_produce_new_form(new_state, form_url=reeval_url)
+        return self._reconcile_state(new_state, form_url=reeval_url)
 
     @raises_locust_error("uiform.py/select_radio_button_by_index()")
     def select_radio_button_by_index(self, field_index: int, index: int, locust_request_label: str = "") -> 'SailUiForm':
@@ -966,7 +1070,7 @@ class SailUiForm:
         if not new_state:
             raise Exception(
                 f"No response returned when trying to select radio button with index '{field_index}'")
-        return self._reconcile_and_produce_new_form(new_state, form_url=reeval_url)
+        return self._reconcile_state(new_state, form_url=reeval_url)
 
     def go_to_next_record_grid_page(self, locust_request_label: str = "") -> 'SailUiForm':
         context_label = locust_request_label or f"{self.breadcrumb}.NextPage"
@@ -978,11 +1082,12 @@ class SailUiForm:
             raise Exception("Not a grid record list")
         component = find_component_by_attribute_in_dict(
             'label', label, self.state)
-        new_state = self.interactor.interact_with_record_grid(post_url=reeval_url, grid_component=component, context=self.context, uuid=self.uuid)
+        new_state = self.interactor.interact_with_record_grid(
+            post_url=reeval_url, grid_component=component, context=self.context, uuid=self.uuid, context_label=context_label)
         if not new_state:
             raise Exception(
                 f"No response returned when navigating to next page on record list '{reeval_url}'")
-        return self._reconcile_and_produce_new_form(new_state, form_url=reeval_url)
+        return self._reconcile_state(new_state, form_url=reeval_url)
 
     def get_record_header_form(self) -> 'SailUiForm':
         """
@@ -1022,6 +1127,29 @@ class SailUiForm:
 
         return SailUiForm(self.interactor, json.loads(record_view_response), self.form_url, breadcrumb=self.breadcrumb)
 
+    def filter_records_using_searchbox(self, search_term: str = "", locust_request_label: str = "") -> 'SailUiForm':
+        """
+        This method allows you to Filter the Record Type List (displaying record instance for a specific record type)
+        which makes the same request when typing something in the search box and reloading the page.
+        More interactions (with the filtered list) can be performed on the returned SailUiForm Object.
+
+        Note: This is different from how an end user interacts with the SearchBox. In that case you would type something in the box
+        and when you would unfocus from the box a reevaluation happens and then you click the 'Search' button which returns the filtered
+        results.
+
+        Examples:
+
+            >>> form.filter_records_using_searchbox('Donuts')
+
+        Returns (SailUiForm): The record type list UiForm with the filtered results.
+        """
+        context_label = locust_request_label or f"{self.breadcrumb}.RecordType.SearchByText"
+        search_uri = f"{self.form_url}?searchTerm={quote(search_term)}"
+
+        headers = self.interactor.setup_sail_headers()
+        response = self.interactor.get_page(uri=search_uri, headers=headers, label=context_label)
+        return SailUiForm(self.interactor, response.json(), self.form_url, breadcrumb=context_label)
+
     def assert_no_validations_present(self) -> 'SailUiForm':
         """
         Raises an exception if there are validations present on the form, otherwise, returns the form as is
@@ -1038,17 +1166,17 @@ class SailUiForm:
             raise Exception(f"At least one validation was found in the form {self.breadcrumb}")
         return self
 
-    def _reconcile_and_produce_new_form(self, new_state: dict, form_url: str = "") -> 'SailUiForm':
-        self.interactor.datatype_cache.cache(new_state)  # Also update datatypes cache
-        reconciled_state = self.reconciler.reconcile_ui(self.state, new_state)
-        new_form_url = form_url if form_url else self.form_url
-        return SailUiForm(self.interactor, reconciled_state, new_form_url, breadcrumb=self.breadcrumb)
+    def _reconcile_state(self, new_state: dict, form_url: str = "") -> 'SailUiForm':
+        self.interactor.datatype_cache.cache(new_state)
+        self.state = self.reconciler.reconcile_ui(self.state, new_state)
+        self.form_url = form_url or self.form_url
+        return self
 
-    def _validate_component_found(self, component: Optional[Dict[str, Any]], label: str) -> None:
+    def _validate_component_found(self, component: Optional[Dict[str, Any]], label: str, type: Optional[str] = None) -> None:
         if not component:
-            raise ComponentNotFoundException(
-                f"Could not find the component with label '{label}' in the provided form"
-            )
+            optional_type_info = f" of type '{type}'" if type else ''
+            msg = f"Could not find the component with label '{label}'{optional_type_info} in the provided form"
+            raise ComponentNotFoundException(msg)
 
     def _validate_component_found_by_attribute(self, component: Dict[str, Any], attribute: str, value_for_attribute: str) -> None:
         if not component:
